<div class="avoidbreak">
  <address class="vcard">
	<span class="vcardline">
	  <span class="fn">$fullname</span> $role
	  <span class="n hidden">
		<span class="family-name">$surname</span>
	  </span>
	</span>
	<span class="org vcardline">$organization</span>
	<span class="adr">
	  $street_spans
	  <span class="vcardline">
		<span class="locality">$locality</span>$locality_sep 
		<span class="region">$region</span>$region_sep
		<span class="code">$code</span>
	  </span>
	  <span class="country-name vcardline">$country</span>
	</span>
	$contact_spans
  </address>
</div>